/*--------------------------------------------------------------------*\
  File:      ProgressBar.cpp
  Creator:   Matt Bogosian <mattb@be.com>
  Copyright: (c)1998, Be, Inc. All rights reserved.
  Description: Source file containing the member functions for the
      ProgressBar class.
\*--------------------------------------------------------------------*/



#include "ProgressBar.h"
#include <Window.h>
#include <string.h>

//====================================================================
ProgressBar::ProgressBar(BRect a_frame, const char * const a_name) :
//====================================================================
	Inherited(a_frame, a_name, B_FOLLOW_NONE, B_WILL_DRAW),
	m_status_bar(NULL),
	m_bitmap(NULL),
	m_delta(0.0),
	m_last_update(0)
{
	a_frame.OffsetTo(0.0, 0.0);
	
	m_bitmap = new BBitmap(a_frame, B_CMAP8, true);
	m_status_bar = new BStatusBar(a_frame, "");
	
	// Set up the attributes which will determine the status bar's
	// height when it's added as a child
	m_status_bar->SetBarHeight(a_frame.Height());
	
	// Add it and remove it to initialize it's real height (we need to
	// do this because Move*() calls have no effect once a view is a
	// child of a bitmap)
	m_bitmap->AddChild(m_status_bar);
	m_bitmap->RemoveChild(m_status_bar);
	
	// Set it up for real
	m_status_bar->MoveTo(0.0, -(m_status_bar->Bounds().Height() - m_status_bar->BarHeight()));
	m_bitmap->AddChild(m_status_bar);
	
	// Draw the initial state
	if (m_bitmap->Lock())
	{
		RedrawStatus();
		m_bitmap->Unlock();
	}
	
	Inherited::SetViewColor(B_TRANSPARENT_COLOR);
}

//====================================================================
ProgressBar::~ProgressBar(void)
//====================================================================
{
	delete m_bitmap;
}

//====================================================================
void ProgressBar::AttachedToWindow(void)
//====================================================================
{
	SetDrawingMode(B_OP_OVER);
}

//====================================================================
void ProgressBar::Draw(BRect a_update_rect)
//====================================================================
{
	DrawBitmap(m_bitmap, a_update_rect, a_update_rect);
}

//====================================================================
void ProgressBar::Reset(void)
//====================================================================
{
	if (m_bitmap->Lock())
	{
		m_status_bar->Reset();
		RedrawStatus();
		m_bitmap->Unlock();
	}
}

//====================================================================
void ProgressBar::SetMaxValue(const float a_val)
//====================================================================
{
	m_status_bar->SetMaxValue(a_val);
}

//====================================================================
void ProgressBar::Update(const float a_delta)
//====================================================================
{
	// Add the new delta to the stored delta and get the current time
	m_delta += a_delta;
	bigtime_t last_update(real_time_clock_usecs());
	
	// Change the status bar if it's been "long" enough
	if ((last_update > m_last_update + 1000000
			|| m_delta / CurrentValue() > 0.01
			&& !(CurrentValue() > MaxValue()
				&& a_delta > 0.0))
		&& m_bitmap->Lock())
	{
		m_status_bar->Update(m_delta);
		RedrawStatus();
		
		// Update the internal state
		m_delta = 0.0;
		m_last_update = last_update;
		
		m_bitmap->Unlock();
		
		// Update the view
		if (Window() != NULL
			&& Window()->Lock())
		{
			Invalidate(Bounds());
			Window()->Unlock();
		}
	}
}

//====================================================================
void ProgressBar::RedrawStatus()
//====================================================================
{
	m_status_bar->PushState();
	const rgb_color high = m_status_bar->HighColor();
	m_status_bar->SetHighColor(m_status_bar->ViewColor());
	m_status_bar->FillRect(m_status_bar->Bounds());
	m_status_bar->SetHighColor(high);
	m_status_bar->Draw(m_status_bar->Bounds());
	m_status_bar->PopState();
	m_status_bar->Sync();
}

const int32 kBevel = 2;
const color_space kBitmapColorSpace = B_COLOR_8_BIT;

extern const unsigned char kBitmapBits1[];
extern const unsigned char kBitmapBits2[];
extern const unsigned char kBitmapBits3[];
extern const unsigned char kBitmapBits4[];
extern const unsigned char kBitmapBits5[];
extern const unsigned char kBitmapBits6[];
extern const unsigned char kBitmapBits7[];

const unsigned char *barberPoleBits[] = {
	kBitmapBits1,
	kBitmapBits2,
	kBitmapBits3,
	kBitmapBits4,
	kBitmapBits5,
	kBitmapBits6,
	kBitmapBits7
};

BarberPoleView::BarberPoleView(BRect frame, const char *name)
	:	BView(frame, name, B_FOLLOW_LEFT + B_FOLLOW_TOP,
			B_PULSE_NEEDED | B_WILL_DRAW),
		bitmap((frame.InsetBySelf(kBevel, kBevel)).OffsetToSelf(0, 0), B_COLOR_8_BIT),
		bits(barberPoleBits),
		count(7), 
		indx(0)
{
}

const rgb_color kLightGray = {150, 150, 150, 255};
const rgb_color kGray = {100, 100, 100, 255};
const rgb_color kWhite = {255, 255, 255, 255};

void 
BarberPoleView::Draw(BRect)
{
	BRect bounds(Bounds());

	BeginLineArray(8);
	AddLine(bounds.LeftTop(), bounds.RightTop(), kLightGray);
	AddLine(bounds.LeftTop(), bounds.LeftBottom(), kLightGray);
	AddLine(bounds.LeftBottom(), bounds.RightBottom(), kWhite);
	AddLine(bounds.RightBottom(), bounds.RightTop(), kWhite);
	bounds.InsetBy(1, 1);
	AddLine(bounds.LeftTop(), bounds.RightTop(), kLightGray);
	AddLine(bounds.LeftTop(), bounds.LeftBottom(), kLightGray);
	AddLine(bounds.LeftBottom(), bounds.RightBottom(), kWhite);
	AddLine(bounds.RightBottom(), bounds.RightTop(), kWhite);
	EndLineArray();
	DrawCurrentBarberPole();
}

void 
BarberPoleView::DrawCurrentBarberPole()
{
	bitmap.SetBits(bits[indx], kBarberPoleBitmapWidth * kBarberPoleBitmapHeight, 0, B_COLOR_8_BIT);
	DrawBitmap(&bitmap, BPoint(kBevel, kBevel));
}


void 
BarberPoleView::Pulse()
{
	DrawCurrentBarberPole();
	if (++indx >= count)
		indx = 0;
}

const unsigned char kBitmapBits1 [] = {
	0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,
	0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,
	0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,
	0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,
	0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,
	0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,
	0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,
	0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,
	0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,
	0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,
	0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,
	0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,
	0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,
	0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,
	0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,
	0x28,0x28,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,
	0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,
	0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,
	0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,
	0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,
	0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,
	0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,
	0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,
	0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,
	0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,
	0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,
	0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,
	0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,
	0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,
	0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,
	0x26,0x28,0x08,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,
	0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,
	0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,
	0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,
	0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,
	0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,
	0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,
	0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,
	0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,
	0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,
	0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,
	0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,
	0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,
	0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,
	0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,
	0x26,0x25,0x28,0x08,0x0f,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x25,0x08,0x08,0x25,0x24,0x23,0x23,0x16,0x17,0x16,0x23,
	0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,
	0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,
	0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,
	0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,
	0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,
	0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,
	0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,
	0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,
	0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,
	0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,
	0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,
	0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,
	0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,
	0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,
	0x23,0x23,0x24,0x0f,0x08,0x28,0x25,0x24,0x23,0x1a,0x1a,0x1a,0x21,
	0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,
	0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,
	0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,
	0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,
	0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,
	0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,
	0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,
	0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,
	0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,
	0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,
	0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,
	0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,
	0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,
	0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,
	0x22,0x22,0x23,0x13,0x0f,0x08,0x28,0x25,0x24,0x17,0x1a,0x1e,0x20,
	0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x22,0x17,0x13,0x0f,0x28,0x28,0x25,0x13,0x17,0x1a,0x20,
	0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,
	0x21,0x21,0x20,0x1a,0x17,0x13,0x25,0x28,0x28,0x0f,0x13,0x16,0x22,
	0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,
	0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,
	0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,
	0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,
	0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,
	0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,
	0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,
	0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,
	0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,
	0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,
	0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,
	0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,
	0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,
	0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,
	0x22,0x22,0x21,0x1a,0x1a,0x16,0x24,0x25,0x28,0x08,0x0f,0x13,0x23,
	0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,
	0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,
	0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,
	0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,
	0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,
	0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,
	0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,
	0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,
	0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,
	0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,
	0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,
	0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,
	0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,
	0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,
	0x23,0x23,0x23,0x16,0x17,0x17,0x23,0x24,0x25,0x28,0x08,0x0f,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x25,0x08,0x08,0x25,
	0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,
	0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,
	0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,
	0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,
	0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,
	0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,
	0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,
	0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,
	0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,
	0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,
	0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,
	0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,
	0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,
	0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,
	0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x08,0x28,
	0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,
	0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,
	0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,
	0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,
	0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,
	0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,
	0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,
	0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,
	0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,
	0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,
	0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,
	0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,
	0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,
	0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,
	0x27,0x27,0x27,0x0b,0x0c,0x0c,0x27,0x27,0x27,0x27,0x0b,0x0c,0x08,
};

const unsigned char kBitmapBits2 [] = {
	0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,
	0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,
	0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,
	0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,
	0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,
	0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,
	0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,
	0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,
	0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,
	0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,
	0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,
	0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,
	0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,
	0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,
	0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,
	0x28,0x28,0x26,0x0c,0x0b,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,
	0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,
	0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,
	0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,
	0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,
	0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,
	0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,
	0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,
	0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,
	0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,
	0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,
	0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,
	0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,
	0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,
	0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,
	0x26,0x28,0x28,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,
	0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,
	0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,
	0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,
	0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,
	0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,
	0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,
	0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,
	0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,
	0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,
	0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,
	0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,
	0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,
	0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,
	0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,
	0x25,0x25,0x28,0x08,0x0f,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x25,0x28,0x08,0x0f,0x24,0x23,0x23,0x23,0x16,0x17,0x16,
	0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,
	0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,
	0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,
	0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,
	0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,
	0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,
	0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,
	0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,
	0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,
	0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,
	0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,
	0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,
	0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,
	0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,
	0x23,0x23,0x24,0x25,0x08,0x08,0x25,0x24,0x23,0x22,0x1a,0x1a,0x1a,
	0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,
	0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,
	0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,
	0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,
	0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,
	0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,
	0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,
	0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,
	0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,
	0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,
	0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,
	0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,
	0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,
	0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,
	0x21,0x22,0x23,0x24,0x0f,0x08,0x28,0x25,0x24,0x23,0x1a,0x1e,0x1e,
	0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,
	0x21,0x20,0x22,0x23,0x13,0x0f,0x08,0x28,0x25,0x24,0x17,0x1a,0x1e,
	0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,
	0x21,0x20,0x21,0x22,0x17,0x13,0x0f,0x28,0x28,0x25,0x13,0x16,0x1a,
	0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,
	0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,
	0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,
	0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,
	0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,
	0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,
	0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,
	0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,
	0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,
	0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,
	0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,
	0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,
	0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,
	0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,
	0x22,0x21,0x22,0x22,0x1b,0x16,0x13,0x25,0x28,0x28,0x0f,0x13,0x17,
	0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,
	0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,
	0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,
	0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,
	0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,
	0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,
	0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,
	0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,
	0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,
	0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,
	0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,
	0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,
	0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,
	0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,
	0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x24,0x25,0x28,0x08,0x0f,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x12,0x24,0x24,0x25,0x28,0x08,0x0f,
	0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,
	0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,
	0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,
	0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,
	0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,
	0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,
	0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,
	0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,
	0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,
	0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,
	0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,
	0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,
	0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,
	0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,
	0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x26,0x25,0x08,0x08,
	0x27,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,
	0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,
	0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,
	0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,
	0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,
	0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,
	0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,
	0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,
	0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,
	0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,
	0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,
	0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,
	0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,
	0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,
	0x26,0x27,0x27,0x27,0x0b,0x0c,0x0c,0x27,0x26,0x27,0x26,0x0b,0x08,
};

const unsigned char kBitmapBits3 [] = {
	0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,
	0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,
	0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,
	0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,
	0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,
	0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,
	0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,
	0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,
	0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,
	0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,
	0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,
	0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,
	0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,
	0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,
	0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,
	0x08,0x28,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,
	0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,
	0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,
	0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,
	0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,
	0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,
	0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,
	0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,
	0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,
	0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,
	0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,
	0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,
	0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,
	0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,
	0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,
	0x0b,0x28,0x28,0x25,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,
	0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,
	0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,
	0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,
	0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,
	0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,
	0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,
	0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,
	0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,
	0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,
	0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,
	0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,
	0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,
	0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,
	0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,
	0x0f,0x25,0x28,0x28,0x0f,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x25,0x28,0x08,0x0f,0x13,0x23,0x23,0x23,0x23,0x16,0x17,
	0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,
	0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,
	0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,
	0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,
	0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,
	0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,
	0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,
	0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,
	0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,
	0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,
	0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,
	0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,
	0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,
	0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,
	0x16,0x23,0x24,0x25,0x28,0x08,0x0f,0x24,0x23,0x22,0x22,0x1b,0x1a,
	0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,
	0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,
	0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,
	0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,
	0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,
	0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,
	0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,
	0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,
	0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,
	0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,
	0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,
	0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,
	0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,
	0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,
	0x1a,0x21,0x23,0x24,0x25,0x08,0x08,0x25,0x24,0x23,0x21,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,
	0x1e,0x20,0x22,0x23,0x24,0x0f,0x08,0x28,0x25,0x24,0x23,0x1a,0x1e,
	0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,
	0x1e,0x20,0x21,0x22,0x23,0x13,0x0f,0x08,0x28,0x25,0x24,0x16,0x1a,
	0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,
	0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,
	0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,
	0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,
	0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,
	0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,
	0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,
	0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,
	0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,
	0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,
	0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,
	0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,
	0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,
	0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,
	0x1a,0x22,0x22,0x22,0x21,0x16,0x13,0x0f,0x28,0x28,0x25,0x13,0x17,
	0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,
	0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,
	0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,
	0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,
	0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,
	0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,
	0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,
	0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,
	0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,
	0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,
	0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,
	0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,
	0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,
	0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,
	0x17,0x23,0x23,0x23,0x23,0x16,0x17,0x13,0x25,0x28,0x28,0x0f,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x25,0x28,0x08,0x0f,
	0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,
	0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,
	0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,
	0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,
	0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,
	0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,
	0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,
	0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,
	0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,
	0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,
	0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,
	0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,
	0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,
	0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,
	0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x28,0x08,
	0x0c,0x27,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,
	0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,
	0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,
	0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,
	0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,
	0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,
	0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,
	0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,
	0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,
	0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,
	0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,
	0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,
	0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,
	0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,
	0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0c,0x27,0x26,0x27,0x27,0x08,
};


const unsigned char kBitmapBits4 [] = {
	0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,
	0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,
	0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,
	0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,
	0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,
	0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,
	0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,
	0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,
	0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,
	0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,
	0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,
	0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,
	0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,
	0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,
	0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,
	0x08,0x28,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,
	0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,
	0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,
	0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,
	0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,
	0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,
	0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,
	0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,
	0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,
	0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,
	0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,
	0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,
	0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,
	0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,
	0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,
	0x0c,0x08,0x28,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,
	0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,
	0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,
	0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,
	0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,
	0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,
	0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,
	0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,
	0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,
	0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,
	0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,
	0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,
	0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,
	0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,
	0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,
	0x0f,0x0f,0x28,0x28,0x25,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x25,0x28,0x28,0x0f,0x13,0x16,0x23,0x23,0x23,0x23,0x16,
	0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,
	0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,
	0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,
	0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,
	0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,
	0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,
	0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,
	0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,
	0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,
	0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,
	0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,
	0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,
	0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,
	0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,
	0x17,0x16,0x24,0x25,0x28,0x08,0x0f,0x13,0x23,0x21,0x22,0x22,0x1b,
	0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,
	0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,
	0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,
	0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,
	0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,
	0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,
	0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,
	0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,
	0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,
	0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,
	0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,
	0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,
	0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,
	0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,
	0x1a,0x1a,0x23,0x24,0x25,0x28,0x08,0x0f,0x24,0x23,0x22,0x20,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,
	0x1e,0x1e,0x22,0x23,0x24,0x25,0x08,0x08,0x25,0x24,0x23,0x22,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,
	0x1e,0x1e,0x21,0x21,0x23,0x24,0x0f,0x08,0x28,0x25,0x24,0x23,0x1a,
	0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,
	0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,
	0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,
	0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,
	0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,
	0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,
	0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,
	0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,
	0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,
	0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,
	0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,
	0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,
	0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,
	0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,
	0x1a,0x1a,0x22,0x21,0x22,0x23,0x13,0x0f,0x08,0x28,0x25,0x24,0x16,
	0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,
	0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,
	0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,
	0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,
	0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,
	0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,
	0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,
	0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,
	0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,
	0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,
	0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,
	0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,
	0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,
	0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,
	0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x13,0x0f,0x28,0x28,0x25,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x25,0x28,0x28,0x0f,
	0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,
	0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,
	0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,
	0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,
	0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,
	0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,
	0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,
	0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,
	0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,
	0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,
	0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,
	0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,
	0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,
	0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,
	0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x28,0x08,
	0x0c,0x0c,0x27,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,
	0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,
	0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,
	0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,
	0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,
	0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,
	0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,
	0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,
	0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,
	0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,
	0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,
	0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,
	0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,
	0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,
	0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x28,
};

const unsigned char kBitmapBits5 [] = {
	0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,
	0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,
	0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,
	0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,
	0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,
	0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,
	0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,
	0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,
	0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,
	0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,
	0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,
	0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,
	0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,
	0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,
	0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,
	0x08,0x08,0x27,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,
	0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,
	0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,
	0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,
	0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,
	0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,
	0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,
	0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,
	0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,
	0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,
	0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,
	0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,
	0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,
	0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,
	0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,
	0x0b,0x08,0x28,0x25,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,
	0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,
	0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,
	0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,
	0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,
	0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,
	0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,
	0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,
	0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,
	0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,
	0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,
	0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,
	0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,
	0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,
	0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,
	0x0f,0x0f,0x08,0x28,0x25,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x0f,0x28,0x28,0x25,0x13,0x16,0x17,0x23,0x23,0x23,0x23,
	0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,
	0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,
	0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,
	0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,
	0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,
	0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,
	0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,
	0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,
	0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,
	0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,
	0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,
	0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,
	0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,
	0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,
	0x16,0x17,0x13,0x25,0x28,0x28,0x0f,0x13,0x17,0x22,0x21,0x22,0x22,
	0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,
	0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,
	0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,
	0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,
	0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,
	0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,
	0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,
	0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,
	0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,
	0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,
	0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,
	0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,
	0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,
	0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,
	0x1b,0x1a,0x17,0x24,0x25,0x28,0x08,0x0f,0x13,0x23,0x22,0x21,0x20,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,
	0x1e,0x1e,0x1a,0x23,0x24,0x25,0x28,0x08,0x0f,0x24,0x23,0x22,0x20,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,
	0x1e,0x1e,0x1e,0x22,0x23,0x24,0x25,0x08,0x08,0x25,0x24,0x23,0x21,
	0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,
	0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,
	0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,
	0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,
	0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,
	0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,
	0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,
	0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,
	0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,
	0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,
	0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,
	0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,
	0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,
	0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,
	0x1a,0x1a,0x1a,0x22,0x21,0x23,0x24,0x0f,0x08,0x28,0x25,0x24,0x23,
	0x16,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,
	0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,
	0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,
	0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,
	0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,
	0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,
	0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,
	0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,
	0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,
	0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,
	0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,
	0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,
	0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,
	0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,
	0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x13,0x0f,0x08,0x28,0x25,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x0f,0x28,0x28,0x25,
	0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,
	0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,
	0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,
	0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,
	0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,
	0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,
	0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,
	0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,
	0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,
	0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,
	0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,
	0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,
	0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,
	0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,
	0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x28,0x28,
	0x0c,0x0b,0x0c,0x27,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,
	0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,
	0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,
	0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,
	0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,
	0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,
	0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,
	0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,
	0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,
	0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,
	0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,
	0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,
	0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,
	0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,
	0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0c,0x27,0x26,0x28,
};


const unsigned char kBitmapBits6 [] = {
	0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,
	0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,
	0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,
	0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,
	0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,
	0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,
	0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,
	0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,
	0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,
	0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,
	0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,
	0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,
	0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,
	0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,
	0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,
	0x28,0x08,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,
	0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,
	0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,
	0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,
	0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,
	0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,
	0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,
	0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,
	0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,
	0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,
	0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,
	0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,
	0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,
	0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,
	0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,
	0x26,0x08,0x08,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,
	0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,
	0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,
	0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,
	0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,
	0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,
	0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,
	0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,
	0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,
	0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,
	0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,
	0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,
	0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,
	0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,
	0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,
	0x26,0x0f,0x08,0x28,0x25,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x0f,0x08,0x28,0x25,0x24,0x16,0x17,0x16,0x23,0x23,0x23,
	0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,
	0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,
	0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,
	0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,
	0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,
	0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,
	0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,
	0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,
	0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,
	0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,
	0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,
	0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,
	0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,
	0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,
	0x23,0x16,0x13,0x0f,0x28,0x28,0x25,0x13,0x17,0x1a,0x22,0x21,0x22,
	0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,
	0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,
	0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,
	0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,
	0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,
	0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,
	0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,
	0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,
	0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,
	0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,
	0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,
	0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,
	0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,
	0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,
	0x22,0x1b,0x17,0x13,0x25,0x28,0x28,0x0f,0x13,0x16,0x22,0x20,0x21,
	0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x20,0x1e,0x1a,0x16,0x24,0x25,0x28,0x08,0x0f,0x13,0x23,0x22,0x20,
	0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x21,0x1e,0x1e,0x1a,0x23,0x24,0x25,0x28,0x08,0x0f,0x24,0x23,0x22,
	0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,
	0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,
	0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,
	0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,
	0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,
	0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,
	0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,
	0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,
	0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,
	0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,
	0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,
	0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,
	0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,
	0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,
	0x22,0x1a,0x1a,0x1a,0x22,0x23,0x24,0x25,0x08,0x08,0x25,0x24,0x23,
	0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,
	0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,
	0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,
	0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,
	0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,
	0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,
	0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,
	0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,
	0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,
	0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,
	0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,
	0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,
	0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,
	0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,
	0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x24,0x0f,0x08,0x28,0x25,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x0f,0x08,0x28,0x25,
	0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,
	0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,
	0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,
	0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,
	0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,
	0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,
	0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,
	0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,
	0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,
	0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,
	0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,
	0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,
	0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,
	0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,
	0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x28,0x28,
	0x26,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,
	0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,
	0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,
	0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,
	0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,
	0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,
	0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,
	0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,
	0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,
	0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,
	0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,
	0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,
	0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,
	0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,
	0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0c,0x26,0x28,
};

const unsigned char kBitmapBits7 [] = {
	0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,
	0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,
	0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,
	0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,
	0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,
	0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,
	0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,
	0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,
	0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,
	0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,
	0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,
	0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,
	0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,
	0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,
	0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,
	0x28,0x08,0x0b,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,
	0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,
	0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,
	0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,
	0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,
	0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,
	0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,
	0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,
	0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,
	0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,
	0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,
	0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,
	0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,
	0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,
	0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,
	0x26,0x28,0x08,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,
	0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,
	0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,
	0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,
	0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,
	0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,
	0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,
	0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,
	0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,
	0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,
	0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,
	0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,
	0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,
	0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,
	0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x26,0x25,
	0x26,0x25,0x08,0x08,0x25,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x0f,0x08,0x28,0x25,0x24,0x23,0x16,0x17,0x16,0x23,0x23,
	0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,
	0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,
	0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,
	0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,
	0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,
	0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,
	0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,
	0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,
	0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,
	0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,
	0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,
	0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,
	0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,
	0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,
	0x23,0x23,0x13,0x0f,0x08,0x28,0x25,0x24,0x16,0x1a,0x1a,0x21,0x22,
	0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,
	0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,
	0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,
	0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,
	0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,
	0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,
	0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,
	0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,
	0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,
	0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,
	0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,
	0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,
	0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,
	0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,
	0x21,0x22,0x17,0x13,0x0f,0x28,0x28,0x25,0x13,0x17,0x1a,0x20,0x21,
	0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x21,0x1a,0x16,0x13,0x25,0x28,0x28,0x0f,0x13,0x17,0x22,0x21,
	0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x21,0x1e,0x1a,0x17,0x24,0x25,0x28,0x08,0x0f,0x13,0x23,0x22,
	0x21,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,
	0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,
	0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,
	0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,
	0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,
	0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,
	0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,
	0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,
	0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,
	0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,
	0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,
	0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,
	0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,
	0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,
	0x22,0x22,0x1a,0x1a,0x1a,0x23,0x24,0x25,0x28,0x08,0x0f,0x24,0x23,
	0x23,0x23,0x16,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,
	0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,
	0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,
	0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,
	0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,
	0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,
	0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,
	0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,
	0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,
	0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,
	0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,
	0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,
	0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,
	0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,
	0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x24,0x25,0x08,0x08,0x25,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x0f,0x08,0x28,0x25,
	0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,
	0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,
	0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,
	0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,
	0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,
	0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,
	0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,
	0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,
	0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,
	0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,
	0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,
	0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,
	0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,
	0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,
	0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x08,0x28,
	0x27,0x27,0x0b,0x0c,0x0c,0x27,0x26,0x27,0x27,0x0b,0x0c,0x0b,0x26,
	0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,
	0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,
	0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,
	0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,
	0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,
	0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,
	0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,
	0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,
	0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,
	0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,
	0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,
	0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,
	0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,
	0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x28,
};

