#include <Bitmap.h>
#include <Box.h>
#include <Button.h>
#include <Debug.h>
#include <StringView.h>

#include "CheckBFS.h"

// offscreen bitmap utility class

OffscreenBitmap::OffscreenBitmap()
	:	bitmap(NULL),
		target(NULL),
		destinationFrame(0, 0, 0, 0)
{
}

OffscreenBitmap::OffscreenBitmap(BView *target, BRect frame)
{
	SetToCommon(target, frame);
}

OffscreenBitmap::~OffscreenBitmap()
{
	delete bitmap;
}

BView * 
OffscreenBitmap::BeginBlitting(BView *newTarget, BRect frame,
	bool copyTargetViewState, bool erase)
{
	if (!bitmap
		|| target != newTarget
		|| frame.OffsetToCopy(0, 0) != destinationFrame.OffsetToCopy(0, 0)) {
		// different view or different bitmap size, just allocate a new bitmap
		delete bitmap;
		SetToCommon(newTarget, frame);
	} else {
		// same bitmap, just make sure it's locked and the origin is
		// set up right
		if (!bitmap->IsLocked() && !bitmap->Lock())
			throw (status_t)B_ERROR;
		destinationFrame = frame;
		if (copyTargetViewState)
			SetUpState(newTarget, frame);
		else
			OffscreenView()->SetOrigin(-frame.left, -frame.top);

		if (erase) 
			OffscreenView()->FillRect(frame, B_SOLID_LOW); 
	}
	return OffscreenView();
}

void 
OffscreenBitmap::SetToCommon(BView *newTarget, BRect frame)
{
	target = newTarget;
	destinationFrame = frame;

	BRect bounds(frame);
	bounds.OffsetTo(0, 0);
	bitmap = new BBitmap(bounds, B_COLOR_8_BIT, true);
	if (!bitmap->Lock())
		throw (status_t)B_ERROR;

	BView *view = new BView(bitmap->Bounds(), "", B_FOLLOW_NONE, 0);
	bitmap->AddChild(view);
	SetUpState(newTarget, frame);
}

void 
OffscreenBitmap::SetUpState(BView *newTarget, BRect frame)
{
	BView *view = bitmap->ChildAt(0);

	// make sure the bitmap matches the target view drawing modes
	view->SetViewColor(newTarget->ViewColor());
	view->SetLowColor(newTarget->LowColor());
	view->SetHighColor(newTarget->HighColor());
	view->SetDrawingMode(newTarget->DrawingMode());
	BFont font;
	newTarget->GetFont(&font);
	view->SetFont(&font);

	view->SetOrigin(-frame.left, -frame.top);
}


BView *
OffscreenBitmap::OffscreenView() const
{
	ASSERT(bitmap);
	return bitmap->ChildAt(0);
}

void 
OffscreenBitmap::DoneBlitting()
{
	// bitmap is composed, output it to the target view
	bitmap->ChildAt(0)->Sync();
	target->SetDrawingMode(B_OP_COPY);
	target->DrawBitmap(bitmap, bitmap->Bounds(), destinationFrame);
	bitmap->Unlock();
}


const BPoint kButtonSize(70, 20);
const uint32 kStop = 'stop';
const uint32 kTextHeight = 18;

const unsigned char *bitmaps[] = {
	kBitmapBits1,
	kBitmapBits2,
	kBitmapBits3,
	kBitmapBits4,
	kBitmapBits5,
	kBitmapBits6,
	kBitmapBits7
};

CheckBFSView::CheckBFSView(BRect frame)
	:	BBox(frame, "", B_FOLLOW_ALL, B_WILL_DRAW | B_PULSE_NEEDED, B_PLAIN_BORDER)
{
	BRect rect(0, 0, kBitmapWidth - 1, kBitmapHeight - 1);
	rect.InsetBy(-2, -2);
	rect.OffsetTo(15, 13);
	barberPole = new BarberPoleView(rect, "barberPole", bitmaps, 7, kDropFilesBitmapBits);
	AddChild(barberPole);

	rect.OffsetBy(0, 15);
	rect.bottom = rect.top + kTextHeight;
	titleText = new BStringView(rect, "titleText", "");
	AddChild(titleText);
	rect.OffsetBy(0, 18);
	progressText = new FlickerFreeStringView(rect, "progressText", "");
	AddChild(progressText);

	rect.OffsetBy(0, 25);
	AddChild(new SeparatorLine(rect.LeftTop(), rect.Width(), false));
	BRect bounds(Bounds());
	rect = bounds;
	rect.InsetBy(12, 12);
	rect.SetLeftTop(rect.RightBottom() - kButtonSize);
	stop = new BButton(rect, "Stop", "Stop", new BMessage(kStop));
	AddChild(stop);

	Done();
}

CheckBFSView::~CheckBFSView()
{
}

void 
CheckBFSView::Pulse()
{
	if (engine.GetCurrentStatus(stateString, progressString)) {
		BString newLabel(stateString);
		float width = progressText->Frame().Width();
		width -= progressText->StringWidth(stateString.String());
		if (width > 0) {
			BString tmp(progressString);
			TruncateString(&tmp, B_TRUNCATE_END, width);
			newLabel += tmp;
		}
		progressText->SetText(newLabel.String());
	}
}


void 
CheckBFSView::AttachedToWindow()
{
	stop->SetTarget(this);
	engine.SetOwner(this);
		// have to set up ownership here because we are using
		// a BMessenger and we cannot build one in the constructor
}

void 
CheckBFSView::MessageReceived(BMessage *message)
{
	switch (message->what) {

		case kStop:
			StopIfNeeded();
			break;

		case kDoneOK:
			Done();
			//if (shouldQuitWhenDone)
			Window()->PostMessage(B_QUIT_REQUESTED);
			break;

		case kDoneError:
			Done();
			break;

		default:
			_inherited::MessageReceived(message);
	}
}

bool 
CheckBFSView::StopIfNeeded()
{
	if (engine.State() != Engine::kBusy && engine.State() != Engine::kPaused)
		return true;

	engine.Pause();
	barberPole->SetPaused();
	Window()->SetFeel(B_NORMAL_WINDOW_FEEL);
		// allow alert to pop up over us
	if (engine.AskToStop()) {
		Stop();
		return true;
	}
	Window()->Activate();
	Window()->SetFeel(B_MODAL_ALL_WINDOW_FEEL);
		// make us modal all again
	engine.Resume();
	barberPole->SetProgressing();
	return false;
}


void 
CheckBFSView::Stop()
{
	barberPole->SetWaitingForDrop();
	engine.Stop();
	Done();
}

void 
CheckBFSView::Done()
{
	barberPole->SetWaitingForDrop();
	stop->SetEnabled(false);
	titleText->SetText("");
}


void 
CheckBFSView::Start(const char *volume)
{
	progressText->SetText("");

	barberPole->SetProgressing();
	stop->SetEnabled(true);
	engine.Start(volume);

	BString text;
	text << "Checking " << volume;
	titleText->SetText(text.String());
}


BRect kWindowRect(0, 0, 230, 110);

CheckBFSWindow::CheckBFSWindow(BPoint offset, const char *volume)
	:	BWindow(kWindowRect.OffsetToSelf(offset), "CheckBFS", B_TITLED_WINDOW_LOOK,
			B_MODAL_ALL_WINDOW_FEEL, B_NOT_ZOOMABLE | B_NOT_RESIZABLE),
		quitting(true)
{
	panel = new CheckBFSView(Bounds());
	AddChild(panel);
	SetPulseRate(200000);
	panel->Start(volume);
}

bool 
CheckBFSWindow::QuitRequested()
{
	if (!panel->StopIfNeeded()
		&& (panel->TargetEngine()->State() == Engine::kBusy
			|| panel->TargetEngine()->State() == Engine::kPaused))
			// user canceled quit
		return false;

	quitting = true;
	be_app->PostMessage(B_QUIT_REQUESTED);
	return true;
}

void 
CheckBFSWindow::MessageReceived(BMessage *message)
{
	switch (message->what) {
		case B_SIMPLE_DATA:
			PostMessage(message, panel);
			break;				
		default:
			_inherited::MessageReceived(message);
	}
}

static bool
CompareColors(const rgb_color a, const rgb_color b)
{
	return a.red == b.red
		&& a.green == b.green
		&& a.blue == b.blue
		&& a.alpha == b.alpha;
}

inline bool
operator==(const rgb_color &a, const rgb_color &b)
{
	return CompareColors(a, b);
}

inline bool
operator!=(const rgb_color &a, const rgb_color &b)
{
	return !CompareColors(a, b);
}

FlickerFreeStringView::FlickerFreeStringView(BRect bounds, const char *name,
	const char *text, uint32 resizeFlags, uint32 flags)
	:	BStringView(bounds, name, text, resizeFlags, flags),
		bitmap(0)
{
}

FlickerFreeStringView::~FlickerFreeStringView()
{
	delete bitmap;
}

void 
FlickerFreeStringView::Draw(BRect frame)
{
	Draw(frame, false);
}

void 
FlickerFreeStringView::Draw(BRect,  bool direct)
{
	BRect bounds(Bounds());
	if (!direct && !bitmap)
		bitmap = new OffscreenBitmap(this, Bounds());
	
	BView *blitView = 0;
	
	if (!direct)
		blitView = bitmap->BeginBlitting(this, bounds);
	else
		blitView = this;

	if (Parent()) {
		viewColor = Parent()->ViewColor();
		lowColor = Parent()->ViewColor();
	}

	BFont font;
	GetFont(&font);

	if (!direct) {
		blitView->SetViewColor(viewColor);
		blitView->SetHighColor(HighColor());
		blitView->SetLowColor(lowColor);

		blitView->SetFont(&font);
	
		blitView->Sync();
	}
	blitView->FillRect(bounds, B_SOLID_LOW);

	if (Text()) {
		BPoint loc;

		font_height	height;
		GetFontHeight(&height);

		edge_info eInfo;
		switch (Alignment()) {
			case B_ALIGN_LEFT:
				{
					// If the first char has a negative left edge give it
					// some more room by shifting that much more to the right.
					font.GetEdges(Text(), 1, &eInfo);
					loc.x = bounds.left + (2 - eInfo.left);
					break;
				}
			case B_ALIGN_CENTER:
				{
					float width = StringWidth(Text());
					float center = (bounds.right - bounds.left) / 2;
					loc.x = center - (width/2);
					break;
				}
			case B_ALIGN_RIGHT:
				{
					float width = StringWidth(Text());
					loc.x = bounds.right - width - 2;
					break;
				}
		}
		loc.y = bounds.bottom - (1 + height.descent);
		blitView->MovePenTo(loc);
		blitView->DrawString(Text());
	}

	if (!direct) 
		bitmap->DoneBlitting();
}

void 
FlickerFreeStringView::AttachedToWindow()
{
	BStringView::AttachedToWindow();
	if (Parent()) {
		viewColor = Parent()->ViewColor();
		lowColor = Parent()->ViewColor();
	}
	SetViewColor(B_TRANSPARENT_32_BIT);
	SetLowColor(B_TRANSPARENT_32_BIT);
}

void 
FlickerFreeStringView::SetViewColor(rgb_color color)
{
	if (viewColor != color) {
		viewColor = color;
		Invalidate();
	}
	BStringView::SetViewColor(B_TRANSPARENT_32_BIT);
}

void 
FlickerFreeStringView::SetLowColor(rgb_color color)
{
	if (lowColor != color) {
		lowColor = color;
		Invalidate();
	}
	BStringView::SetLowColor(B_TRANSPARENT_32_BIT);
}

static BRect
LineBounds(BPoint where, float length, bool vertical)
{
	BRect result;
	result.SetLeftTop(where);
	result.SetRightBottom(where + BPoint(2, 2));
	if (vertical)
		result.bottom = result.top + length;
	else
		result.right = result.left + length;
	
	return result;
}

SeparatorLine::SeparatorLine(BPoint where, float length, bool vertical, const char *name)
	: 	BView(LineBounds(where, length, vertical), name,
			B_FOLLOW_LEFT | B_FOLLOW_TOP, B_WILL_DRAW)
{
	SetViewColor(ui_color(B_PANEL_BACKGROUND_COLOR));
	SetLowColor(ui_color(B_PANEL_BACKGROUND_COLOR));
}

const rgb_color kLightGray = {150, 150, 150, 255};
const rgb_color kGray = {100, 100, 100, 255};
const rgb_color kWhite = {255, 255, 255, 255};

void
SeparatorLine::Draw(BRect)
{
	BRect bounds(Bounds());
	rgb_color hiliteColor = tint_color(ViewColor(), 1.5);

	bool vertical = (bounds.left > bounds.right - 3);
	BeginLineArray(2);
	if (vertical) {
		AddLine(bounds.LeftTop(), bounds.LeftBottom(), hiliteColor);
		AddLine(bounds.LeftTop() + BPoint(1, 0), bounds.LeftBottom() + BPoint(1, 0), kWhite);
	} else {
		AddLine(bounds.LeftTop(), bounds.RightTop(), hiliteColor);
		AddLine(bounds.LeftTop() + BPoint(0, 1), bounds.RightTop() + BPoint(0, 1), kWhite);
	}
	EndLineArray();
}

BarberPoleView::BarberPoleView(BRect frame, const char *name,
	const unsigned char *const *barberPoleArray,
	int32 barberPoleCount,
	const unsigned char *dropHereBits)
	:	BView(frame, name, B_FOLLOW_LEFT + B_FOLLOW_TOP,
			B_PULSE_NEEDED | B_WILL_DRAW),
		bitmap((frame.InsetBySelf(kBevel, kBevel)).OffsetToSelf(0, 0), B_COLOR_8_BIT),
		bits(barberPoleArray),
		dropHereBits(dropHereBits),
		count(barberPoleCount), 
		indx(0),
		progress(true),
		paused(false)
{
}

void 
BarberPoleView::Draw(BRect)
{
	BRect bounds(Bounds());

	BeginLineArray(8);
	AddLine(bounds.LeftTop(), bounds.RightTop(), kLightGray);
	AddLine(bounds.LeftTop(), bounds.LeftBottom(), kLightGray);
	AddLine(bounds.LeftBottom(), bounds.RightBottom(), kWhite);
	AddLine(bounds.RightBottom(), bounds.RightTop(), kWhite);
	bounds.InsetBy(1, 1);
	AddLine(bounds.LeftTop(), bounds.RightTop(), kLightGray);
	AddLine(bounds.LeftTop(), bounds.LeftBottom(), kLightGray);
	AddLine(bounds.LeftBottom(), bounds.RightBottom(), kWhite);
	AddLine(bounds.RightBottom(), bounds.RightTop(), kWhite);
	EndLineArray();
	DrawCurrentBarberPole();
}

void 
BarberPoleView::DrawCurrentBarberPole()
{
	if (progress)
		bitmap.SetBits(bits[indx], kBitmapWidth * kBitmapHeight, 0, B_COLOR_8_BIT);
	else
		bitmap.SetBits(dropHereBits, kBitmapWidth * kBitmapHeight, 0, B_COLOR_8_BIT);

	DrawBitmap(&bitmap, BPoint(kBevel, kBevel));
}

void 
BarberPoleView::Pulse()
{
	if (!progress || paused)
		return;

	DrawCurrentBarberPole();
	if (++indx >= count)
		indx = 0;
}

void 
BarberPoleView::SetPaused()
{
	progress = true;
	paused = true;
	Invalidate();
}

void 
BarberPoleView::SetProgressing()
{
	progress = true;
	paused = false;
	Invalidate();
}

void 
BarberPoleView::SetWaitingForDrop()
{
	progress = false;
	paused = true;
	Invalidate();
}


CheckBFSApp::CheckBFSApp()
	:	BApplication("application/x-vnd.Be.CheckBFS")
{
}

void 
CheckBFSApp::MessageReceived(BMessage *message)
{
	switch (message->what) {
		default:
			_inherited::MessageReceived(message);
			break;		
	}
}

void 
CheckBFSApp::ArgvReceived(int32, char **argv)
{
	++argv;
	if (*argv)
		volume << *argv;
}

void 
CheckBFSApp::ReadyToRun()
{
	if (!volume.Length())
		 volume = "boot";

	(new CheckBFSWindow(BPoint(100, 100), volume.String()))->Show();
}

int
main()
{
	CheckBFSApp app;
	app.Run();
	
	return 0;
}


const unsigned char kBitmapBits1 [] = {
	0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,
	0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,
	0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,
	0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,
	0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,
	0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,
	0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,
	0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,
	0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,
	0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,
	0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,
	0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,
	0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,
	0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,
	0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,
	0x28,0x28,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,
	0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,
	0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,
	0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,
	0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,
	0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,
	0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,
	0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,
	0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,
	0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,
	0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,
	0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,
	0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,
	0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,
	0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,
	0x26,0x28,0x08,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,
	0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,
	0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,
	0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,
	0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,
	0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,
	0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,
	0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,
	0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,
	0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,
	0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,
	0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,
	0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,
	0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,
	0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,
	0x26,0x25,0x28,0x08,0x0f,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x25,0x08,0x08,0x25,0x24,0x23,0x23,0x16,0x17,0x16,0x23,
	0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,
	0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,
	0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,
	0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,
	0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,
	0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,
	0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,
	0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,
	0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,
	0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,
	0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,
	0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,
	0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,
	0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,
	0x23,0x23,0x24,0x0f,0x08,0x28,0x25,0x24,0x23,0x1a,0x1a,0x1a,0x21,
	0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,
	0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,
	0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,
	0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,
	0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,
	0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,
	0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,
	0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,
	0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,
	0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,
	0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,
	0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,
	0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,
	0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,
	0x22,0x22,0x23,0x13,0x0f,0x08,0x28,0x25,0x24,0x17,0x1a,0x1e,0x20,
	0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x22,0x17,0x13,0x0f,0x28,0x28,0x25,0x13,0x17,0x1a,0x20,
	0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,
	0x21,0x21,0x20,0x1a,0x17,0x13,0x25,0x28,0x28,0x0f,0x13,0x16,0x22,
	0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,
	0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,
	0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,
	0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,
	0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,
	0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,
	0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,
	0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,
	0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,
	0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,
	0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,
	0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,
	0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,
	0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,
	0x22,0x22,0x21,0x1a,0x1a,0x16,0x24,0x25,0x28,0x08,0x0f,0x13,0x23,
	0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,
	0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,
	0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,
	0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,
	0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,
	0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,
	0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,
	0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,
	0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,
	0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,
	0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,
	0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,
	0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,
	0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,
	0x23,0x23,0x23,0x16,0x17,0x17,0x23,0x24,0x25,0x28,0x08,0x0f,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x25,0x08,0x08,0x25,
	0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,
	0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,
	0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,
	0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,
	0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,
	0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,
	0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,
	0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,
	0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,
	0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,
	0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,
	0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,
	0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,
	0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,
	0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x08,0x28,
	0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,
	0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,
	0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,
	0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,
	0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,
	0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,
	0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,
	0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,
	0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,
	0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,
	0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,
	0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,
	0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,
	0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,
	0x27,0x27,0x27,0x0b,0x0c,0x0c,0x27,0x27,0x27,0x27,0x0b,0x0c,0x08,
};

const unsigned char kBitmapBits2 [] = {
	0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,
	0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,
	0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,
	0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,
	0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,
	0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,
	0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,
	0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,
	0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,
	0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,
	0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,
	0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,
	0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,
	0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,
	0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,
	0x28,0x28,0x26,0x0c,0x0b,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,
	0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,
	0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,
	0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,
	0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,
	0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,
	0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,
	0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,
	0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,
	0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,
	0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,
	0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,
	0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,
	0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,
	0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,
	0x26,0x28,0x28,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,
	0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,
	0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,
	0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,
	0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,
	0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,
	0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,
	0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,
	0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,
	0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,
	0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,
	0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,
	0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,
	0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,
	0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,
	0x25,0x25,0x28,0x08,0x0f,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x25,0x28,0x08,0x0f,0x24,0x23,0x23,0x23,0x16,0x17,0x16,
	0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,
	0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,
	0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,
	0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,
	0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,
	0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,
	0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,
	0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,
	0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,
	0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,
	0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,
	0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,
	0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,
	0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,
	0x23,0x23,0x24,0x25,0x08,0x08,0x25,0x24,0x23,0x22,0x1a,0x1a,0x1a,
	0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,
	0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,
	0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,
	0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,
	0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,
	0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,
	0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,
	0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,
	0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,
	0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,
	0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,
	0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,
	0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,
	0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,
	0x21,0x22,0x23,0x24,0x0f,0x08,0x28,0x25,0x24,0x23,0x1a,0x1e,0x1e,
	0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,
	0x21,0x20,0x22,0x23,0x13,0x0f,0x08,0x28,0x25,0x24,0x17,0x1a,0x1e,
	0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,
	0x21,0x20,0x21,0x22,0x17,0x13,0x0f,0x28,0x28,0x25,0x13,0x16,0x1a,
	0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,
	0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,
	0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,
	0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,
	0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,
	0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,
	0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,
	0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,
	0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,
	0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,
	0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,
	0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,
	0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,
	0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,
	0x22,0x21,0x22,0x22,0x1b,0x16,0x13,0x25,0x28,0x28,0x0f,0x13,0x17,
	0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,
	0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,
	0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,
	0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,
	0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,
	0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,
	0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,
	0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,
	0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,
	0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,
	0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,
	0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,
	0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,
	0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,
	0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x24,0x25,0x28,0x08,0x0f,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x12,0x24,0x24,0x25,0x28,0x08,0x0f,
	0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,
	0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,
	0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,
	0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,
	0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,
	0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,
	0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,
	0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,
	0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,
	0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,
	0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,
	0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,
	0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,
	0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,
	0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x26,0x25,0x08,0x08,
	0x27,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,
	0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,
	0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,
	0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,
	0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,
	0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,
	0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,
	0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,
	0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,
	0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,
	0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,
	0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,
	0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,
	0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,
	0x26,0x27,0x27,0x27,0x0b,0x0c,0x0c,0x27,0x26,0x27,0x26,0x0b,0x08,
};

const unsigned char kBitmapBits3 [] = {
	0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,
	0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,
	0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,
	0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,
	0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,
	0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,
	0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,
	0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,
	0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,
	0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,
	0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,
	0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,
	0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,
	0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,
	0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,
	0x08,0x28,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,
	0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,
	0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,
	0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,
	0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,
	0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,
	0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,
	0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,
	0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,
	0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,
	0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,
	0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,
	0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,
	0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,
	0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,
	0x0b,0x28,0x28,0x25,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,
	0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,
	0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,
	0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,
	0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,
	0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,
	0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,
	0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,
	0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,
	0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,
	0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,
	0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,
	0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,
	0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,
	0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,
	0x0f,0x25,0x28,0x28,0x0f,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x25,0x28,0x08,0x0f,0x13,0x23,0x23,0x23,0x23,0x16,0x17,
	0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,
	0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,
	0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,
	0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,
	0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,
	0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,
	0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,
	0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,
	0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,
	0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,
	0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,
	0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,
	0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,
	0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,
	0x16,0x23,0x24,0x25,0x28,0x08,0x0f,0x24,0x23,0x22,0x22,0x1b,0x1a,
	0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,
	0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,
	0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,
	0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,
	0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,
	0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,
	0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,
	0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,
	0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,
	0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,
	0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,
	0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,
	0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,
	0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,
	0x1a,0x21,0x23,0x24,0x25,0x08,0x08,0x25,0x24,0x23,0x21,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,
	0x1e,0x20,0x22,0x23,0x24,0x0f,0x08,0x28,0x25,0x24,0x23,0x1a,0x1e,
	0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,
	0x1e,0x20,0x21,0x22,0x23,0x13,0x0f,0x08,0x28,0x25,0x24,0x16,0x1a,
	0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,
	0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,
	0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,
	0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,
	0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,
	0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,
	0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,
	0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,
	0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,
	0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,
	0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,
	0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,
	0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,
	0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,
	0x1a,0x22,0x22,0x22,0x21,0x16,0x13,0x0f,0x28,0x28,0x25,0x13,0x17,
	0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,
	0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,
	0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,
	0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,
	0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,
	0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,
	0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,
	0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,
	0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,
	0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,
	0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,
	0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,
	0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,
	0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,
	0x17,0x23,0x23,0x23,0x23,0x16,0x17,0x13,0x25,0x28,0x28,0x0f,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x25,0x28,0x08,0x0f,
	0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,
	0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,
	0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,
	0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,
	0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,
	0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,
	0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,
	0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,
	0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,
	0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,
	0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,
	0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,
	0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,
	0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,
	0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x28,0x08,
	0x0c,0x27,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,
	0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,
	0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,
	0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,
	0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,
	0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,
	0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,
	0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,
	0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,
	0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,
	0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,
	0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,
	0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,
	0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,
	0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0c,0x27,0x26,0x27,0x27,0x08,
};


const unsigned char kBitmapBits4 [] = {
	0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,
	0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,
	0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,
	0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,
	0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,
	0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,
	0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,
	0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,
	0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,
	0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,
	0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,
	0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,
	0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,
	0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,
	0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,
	0x08,0x28,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,
	0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,
	0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,
	0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,
	0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,
	0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,
	0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,
	0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,
	0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,
	0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,
	0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,
	0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,
	0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,
	0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,
	0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,
	0x0c,0x08,0x28,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,
	0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,
	0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,
	0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,
	0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,
	0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,
	0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,
	0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,
	0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,
	0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,
	0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,
	0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,
	0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,
	0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,
	0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,
	0x0f,0x0f,0x28,0x28,0x25,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x25,0x28,0x28,0x0f,0x13,0x16,0x23,0x23,0x23,0x23,0x16,
	0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,
	0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,
	0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,
	0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,
	0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,
	0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,
	0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,
	0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,
	0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,
	0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,
	0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,
	0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,
	0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,
	0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,
	0x17,0x16,0x24,0x25,0x28,0x08,0x0f,0x13,0x23,0x21,0x22,0x22,0x1b,
	0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,
	0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,
	0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,
	0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,
	0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,
	0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,
	0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,
	0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,
	0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,
	0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,
	0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,
	0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,
	0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,
	0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,
	0x1a,0x1a,0x23,0x24,0x25,0x28,0x08,0x0f,0x24,0x23,0x22,0x20,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,
	0x1e,0x1e,0x22,0x23,0x24,0x25,0x08,0x08,0x25,0x24,0x23,0x22,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,
	0x1e,0x1e,0x21,0x21,0x23,0x24,0x0f,0x08,0x28,0x25,0x24,0x23,0x1a,
	0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,
	0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,
	0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,
	0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,
	0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,
	0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,
	0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,
	0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,
	0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,
	0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,
	0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,
	0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,
	0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,
	0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,
	0x1a,0x1a,0x22,0x21,0x22,0x23,0x13,0x0f,0x08,0x28,0x25,0x24,0x16,
	0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,
	0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,
	0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,
	0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,
	0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,
	0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,
	0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,
	0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,
	0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,
	0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,
	0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,
	0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,
	0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,
	0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,
	0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x13,0x0f,0x28,0x28,0x25,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x25,0x28,0x28,0x0f,
	0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,
	0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,
	0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,
	0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,
	0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,
	0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,
	0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,
	0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,
	0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,
	0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,
	0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,
	0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,
	0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,
	0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,
	0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x28,0x08,
	0x0c,0x0c,0x27,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,
	0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,
	0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,
	0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,
	0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,
	0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,
	0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,
	0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,
	0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,
	0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,
	0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,
	0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,
	0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,
	0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,
	0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x28,
};

const unsigned char kBitmapBits5 [] = {
	0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,
	0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,
	0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,
	0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,
	0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,
	0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,
	0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,
	0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,
	0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,
	0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,
	0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,
	0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,
	0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,
	0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,
	0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,
	0x08,0x08,0x27,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,
	0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,
	0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,
	0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,
	0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,
	0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,
	0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,
	0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,
	0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,
	0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,
	0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,
	0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,
	0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,
	0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,
	0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,
	0x0b,0x08,0x28,0x25,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,
	0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,
	0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,
	0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,
	0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,
	0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,
	0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,
	0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,
	0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,
	0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,
	0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,
	0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,
	0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,
	0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,
	0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,
	0x0f,0x0f,0x08,0x28,0x25,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x0f,0x28,0x28,0x25,0x13,0x16,0x17,0x23,0x23,0x23,0x23,
	0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,
	0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,
	0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,
	0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,
	0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,
	0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,
	0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,
	0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,
	0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,
	0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,
	0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,
	0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,
	0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,
	0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,
	0x16,0x17,0x13,0x25,0x28,0x28,0x0f,0x13,0x17,0x22,0x21,0x22,0x22,
	0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,
	0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,
	0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,
	0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,
	0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,
	0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,
	0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,
	0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,
	0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,
	0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,
	0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,
	0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,
	0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,
	0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,
	0x1b,0x1a,0x17,0x24,0x25,0x28,0x08,0x0f,0x13,0x23,0x22,0x21,0x20,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,
	0x1e,0x1e,0x1a,0x23,0x24,0x25,0x28,0x08,0x0f,0x24,0x23,0x22,0x20,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,
	0x1e,0x1e,0x1e,0x22,0x23,0x24,0x25,0x08,0x08,0x25,0x24,0x23,0x21,
	0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,
	0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,
	0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,
	0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,
	0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,
	0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,
	0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,
	0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,
	0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,
	0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,
	0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,
	0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,
	0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,
	0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,
	0x1a,0x1a,0x1a,0x22,0x21,0x23,0x24,0x0f,0x08,0x28,0x25,0x24,0x23,
	0x16,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,
	0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,
	0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,
	0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,
	0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,
	0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,
	0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,
	0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,
	0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,
	0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,
	0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,
	0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,
	0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,
	0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,
	0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x13,0x0f,0x08,0x28,0x25,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x0f,0x28,0x28,0x25,
	0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,
	0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,
	0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,
	0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,
	0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,
	0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,
	0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,
	0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,
	0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,
	0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,
	0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,
	0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,
	0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,
	0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,
	0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x28,0x28,
	0x0c,0x0b,0x0c,0x27,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,
	0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,
	0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,
	0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,
	0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,
	0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,
	0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,
	0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,
	0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,
	0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,
	0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,
	0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,
	0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,
	0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,
	0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0c,0x27,0x26,0x28,
};


const unsigned char kBitmapBits6 [] = {
	0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,
	0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,
	0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,
	0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,
	0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,
	0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,
	0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,
	0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,
	0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,
	0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,
	0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,
	0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,
	0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,
	0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,
	0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,
	0x28,0x08,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,
	0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,
	0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,
	0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,
	0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,
	0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,
	0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,
	0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,
	0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,
	0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,
	0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,
	0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,
	0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,
	0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,
	0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,
	0x26,0x08,0x08,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,
	0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,
	0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,
	0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,
	0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,
	0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,
	0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,
	0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,
	0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,
	0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,
	0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,
	0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,
	0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,
	0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,
	0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,
	0x26,0x0f,0x08,0x28,0x25,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x0f,0x08,0x28,0x25,0x24,0x16,0x17,0x16,0x23,0x23,0x23,
	0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,
	0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,
	0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,
	0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,
	0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,
	0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,
	0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,
	0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,
	0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,
	0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,
	0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,
	0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,
	0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,
	0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,
	0x23,0x16,0x13,0x0f,0x28,0x28,0x25,0x13,0x17,0x1a,0x22,0x21,0x22,
	0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,
	0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,
	0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,
	0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,
	0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,
	0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,
	0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,
	0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,
	0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,
	0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,
	0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,
	0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,
	0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,
	0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,
	0x22,0x1b,0x17,0x13,0x25,0x28,0x28,0x0f,0x13,0x16,0x22,0x20,0x21,
	0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x20,0x1e,0x1a,0x16,0x24,0x25,0x28,0x08,0x0f,0x13,0x23,0x22,0x20,
	0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x21,0x1e,0x1e,0x1a,0x23,0x24,0x25,0x28,0x08,0x0f,0x24,0x23,0x22,
	0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,
	0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,
	0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,
	0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,
	0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,
	0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,
	0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,
	0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,
	0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,
	0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,
	0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,
	0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,
	0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,
	0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,
	0x22,0x1a,0x1a,0x1a,0x22,0x23,0x24,0x25,0x08,0x08,0x25,0x24,0x23,
	0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,
	0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,
	0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,
	0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,
	0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,
	0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,
	0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,
	0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,
	0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,
	0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,
	0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,
	0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,
	0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,
	0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,
	0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x24,0x0f,0x08,0x28,0x25,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x0f,0x08,0x28,0x25,
	0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,
	0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,
	0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,
	0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,
	0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,
	0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,
	0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,
	0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,
	0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,
	0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,
	0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,
	0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,
	0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,
	0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,
	0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x28,0x28,
	0x26,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,
	0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,
	0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,
	0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,
	0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,
	0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,
	0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,
	0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,
	0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,
	0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,
	0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,
	0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,
	0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,
	0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,
	0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0c,0x26,0x28,
};

const unsigned char kBitmapBits7 [] = {
	0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,
	0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,
	0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,
	0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,
	0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,
	0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,
	0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,
	0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,
	0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,
	0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,
	0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,
	0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,
	0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,
	0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,0x28,
	0x28,0x08,0x08,0x08,0x28,0x28,0x28,0x28,0x08,0x08,0x08,0x28,0x28,
	0x28,0x08,0x0b,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,
	0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,
	0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,
	0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,
	0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,
	0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,
	0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,
	0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,
	0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,
	0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,
	0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,
	0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,
	0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,
	0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,
	0x26,0x0b,0x0c,0x0b,0x26,0x26,0x26,0x26,0x0b,0x0c,0x0b,0x26,0x26,
	0x26,0x28,0x08,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,
	0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,
	0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,
	0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,
	0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,
	0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,
	0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,
	0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,
	0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,
	0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,
	0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,
	0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,
	0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,
	0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,
	0x26,0x0f,0x0f,0x0f,0x25,0x26,0x25,0x26,0x0f,0x0f,0x0f,0x26,0x25,
	0x26,0x25,0x08,0x08,0x25,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x0f,0x08,0x28,0x25,0x24,0x23,0x16,0x17,0x16,0x23,0x23,
	0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,
	0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,
	0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,
	0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,
	0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,
	0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,
	0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,
	0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,
	0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,
	0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,
	0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,
	0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,
	0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,0x23,
	0x23,0x16,0x17,0x16,0x23,0x23,0x23,0x23,0x16,0x17,0x16,0x23,0x23,
	0x23,0x23,0x13,0x0f,0x08,0x28,0x25,0x24,0x16,0x1a,0x1a,0x21,0x22,
	0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,
	0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,
	0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,
	0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,
	0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,
	0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,
	0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,
	0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,
	0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,
	0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,
	0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,
	0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,
	0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,
	0x22,0x1b,0x1a,0x1a,0x21,0x22,0x22,0x22,0x1b,0x1a,0x1a,0x21,0x22,
	0x21,0x22,0x17,0x13,0x0f,0x28,0x28,0x25,0x13,0x17,0x1a,0x20,0x21,
	0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x20,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x20,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x21,0x1a,0x16,0x13,0x25,0x28,0x28,0x0f,0x13,0x17,0x22,0x21,
	0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,
	0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,
	0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,
	0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,
	0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,
	0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,
	0x21,0x1e,0x1e,0x1e,0x20,0x21,0x20,0x21,0x1e,0x1e,0x1e,0x20,0x21,
	0x20,0x21,0x1e,0x1a,0x17,0x24,0x25,0x28,0x08,0x0f,0x13,0x23,0x22,
	0x21,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,
	0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,
	0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,
	0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,
	0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,
	0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,
	0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,
	0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,
	0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,
	0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,
	0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,
	0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,
	0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,
	0x22,0x1a,0x1a,0x1a,0x22,0x22,0x22,0x22,0x1a,0x1a,0x1a,0x22,0x22,
	0x22,0x22,0x1a,0x1a,0x1a,0x23,0x24,0x25,0x28,0x08,0x0f,0x24,0x23,
	0x23,0x23,0x16,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,
	0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,
	0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,
	0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,
	0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,
	0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,
	0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,
	0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,
	0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,
	0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,
	0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,
	0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,
	0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x23,
	0x23,0x17,0x17,0x17,0x23,0x23,0x23,0x23,0x17,0x17,0x17,0x23,0x23,
	0x23,0x23,0x17,0x17,0x17,0x23,0x23,0x24,0x25,0x08,0x08,0x25,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,
	0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,
	0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,
	0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,
	0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,
	0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,
	0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x13,0x13,0x13,0x24,0x24,
	0x24,0x24,0x13,0x13,0x13,0x24,0x24,0x24,0x24,0x0f,0x08,0x28,0x25,
	0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,
	0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,
	0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,
	0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,
	0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,
	0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,
	0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,
	0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,
	0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,
	0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,
	0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,
	0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,
	0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,
	0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,
	0x25,0x25,0x0f,0x0f,0x0f,0x25,0x25,0x25,0x25,0x0f,0x0f,0x08,0x28,
	0x27,0x27,0x0b,0x0c,0x0c,0x27,0x26,0x27,0x27,0x0b,0x0c,0x0b,0x26,
	0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,
	0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,
	0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,
	0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,
	0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,
	0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,
	0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,
	0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,
	0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,
	0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,
	0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,
	0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,
	0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,
	0x27,0x27,0x0b,0x0c,0x0b,0x26,0x27,0x27,0x27,0x0b,0x0c,0x0b,0x28,
};

const unsigned char kDropFilesBitmapBits [] = {
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,
	0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,
	0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,
	0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,
	0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x13,0x13,
	0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,
	0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,
	0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,
	0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,
	0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,
	0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,
	0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,
	0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x13,0x13,
	0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,
	0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,
	0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,
	0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,
	0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,
	0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,
	0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,
	0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x16,0x1c,0x1b,0x1a,0x1b,0x0f,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x17,0x1c,0x0c,0x1b,0x0c,0x1a,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0e,0x11,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x11,0x16,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x13,0x13,
	0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,
	0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,
	0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,
	0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,
	0x0c,0x0c,0x13,0x0c,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,
	0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,
	0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,
	0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,
	0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x16,0x11,0x0c,0x0c,0x10,0x1c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x1a,0x0c,0x0c,0x0c,0x0c,0x1b,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x11,0x16,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x13,
	0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,
	0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,
	0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,
	0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,
	0x0c,0x0c,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,
	0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,
	0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,
	0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,
	0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x16,0x11,0x0c,0x0c,0x0c,0x18,0x0f,0x16,0x18,0x3f,
	0x0d,0x1b,0x1a,0x1c,0x13,0x0c,0x16,0x18,0x1b,0x1c,0x13,0x0c,0x0c,
	0x0c,0x13,0x3f,0x1a,0x0c,0x1b,0x0c,0x1b,0x0c,0x0f,0x1c,0x1b,0x19,
	0x0d,0x0f,0x1c,0x1b,0x1c,0x0e,0x0c,0x0c,0x0c,0x1c,0x1d,0x13,0x10,
	0x1c,0x1a,0x1c,0x0f,0x0c,0x0c,0x0c,0x0f,0x1b,0x1a,0x1d,0x16,0x11,
	0x16,0x0c,0x1c,0x19,0x1b,0x1a,0x0d,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,
	0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,
	0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,
	0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,
	0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,
	0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,
	0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,
	0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,
	0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x16,0x11,0x0c,0x0c,0x0c,0x16,0x11,0x16,0x18,
	0x0c,0x17,0x11,0x0c,0x0c,0x1b,0x0d,0x16,0x16,0x0c,0x0c,0x1b,0x0d,
	0x0c,0x0c,0x0c,0x1a,0x0c,0x0c,0x1b,0x0c,0x1a,0x0c,0x1b,0x0d,0x0c,
	0x10,0x18,0x16,0x13,0x0c,0x10,0x11,0x0c,0x0c,0x0c,0x11,0x16,0x0c,
	0x1b,0x0d,0x0c,0x0d,0x1b,0x0c,0x0c,0x0c,0x0c,0x0c,0x10,0x1d,0x0d,
	0x11,0x16,0x0c,0x1e,0x0d,0x0c,0x11,0x17,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,
	0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,
	0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,
	0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,
	0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,
	0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,
	0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,
	0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,
	0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x16,0x11,0x0c,0x0c,0x0c,0x17,0x10,0x16,
	0x11,0x0c,0x1b,0x0c,0x0c,0x0c,0x16,0x11,0x16,0x11,0x0c,0x0c,0x16,
	0x11,0x0c,0x0c,0x0c,0x1b,0x0c,0x0c,0x1b,0x0c,0x1b,0x0c,0x1e,0x1b,
	0x1b,0x1b,0x1b,0x0d,0x1a,0x1d,0x19,0x0f,0x0c,0x0c,0x0c,0x11,0x16,
	0x0c,0x1b,0x0c,0x0c,0x0c,0x1b,0x0c,0x0c,0x0c,0x0c,0x0d,0x1c,0x10,
	0x0c,0x11,0x16,0x0c,0x1b,0x0c,0x0c,0x0c,0x1b,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,
	0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,
	0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,
	0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,
	0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x0c,0x0c,
	0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,
	0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,
	0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,
	0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,
	0x13,0x0c,0x0c,0x0c,0x0c,0x0c,0x16,0x11,0x0c,0x0c,0x0d,0x1b,0x0c,
	0x16,0x11,0x0c,0x18,0x0e,0x0c,0x0c,0x18,0x0e,0x16,0x13,0x0c,0x0c,
	0x18,0x0e,0x0c,0x0c,0x0c,0x1b,0x0c,0x0c,0x1a,0x0c,0x1b,0x0c,0x1c,
	0x0c,0x0c,0x0c,0x10,0x13,0x0c,0x0c,0x10,0x1b,0x0c,0x0c,0x0c,0x11,
	0x16,0x0c,0x1b,0x0c,0x0c,0x0c,0x1a,0x0c,0x0c,0x0c,0x0c,0x19,0x15,
	0x0c,0x0c,0x11,0x16,0x0c,0x1d,0x0c,0x0c,0x0e,0x18,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,
	0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,
	0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,
	0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,
	0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x0c,0x13,0x0c,0x0c,
	0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,
	0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,
	0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,
	0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,
	0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x0c,0x16,0x18,0x16,0x16,0x1c,0x0f,
	0x0c,0x16,0x11,0x0c,0x10,0x1c,0x16,0x18,0x18,0x0c,0x16,0x1b,0x17,
	0x18,0x18,0x0c,0x0c,0x0c,0x0c,0x1a,0x0c,0x0c,0x1b,0x0c,0x1b,0x0c,
	0x13,0x1b,0x16,0x1a,0x13,0x14,0x1b,0x16,0x19,0x13,0x0c,0x0c,0x0c,
	0x0f,0x1b,0x11,0x13,0x1b,0x16,0x1b,0x13,0x0c,0x0c,0x0c,0x14,0x3f,
	0x16,0x16,0x13,0x11,0x16,0x0c,0x1d,0x1a,0x16,0x1c,0x10,0x0c,0x17,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,
	0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,
	0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,
	0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,
	0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,
	0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,
	0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,
	0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,
	0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0e,0x11,0x11,0x11,0x0c,
	0x0c,0x0c,0x0e,0x0d,0x0c,0x0c,0x0d,0x11,0x10,0x0c,0x0c,0x16,0x11,
	0x10,0x10,0x0c,0x0c,0x0c,0x0c,0x0c,0x10,0x0c,0x0c,0x0f,0x0c,0x0f,
	0x0c,0x0c,0x0e,0x11,0x0d,0x0c,0x0c,0x0e,0x11,0x0e,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0f,0x0e,0x0c,0x0e,0x11,0x0e,0x0c,0x0c,0x0c,0x0c,0x0e,
	0x11,0x11,0x11,0x0f,0x0d,0x0e,0x0c,0x1b,0x0d,0x11,0x0d,0x0c,0x0c,
	0x0f,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,
	0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,
	0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,
	0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,
	0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,
	0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,
	0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,
	0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,
	0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x16,
	0x11,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x1b,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,
	0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,
	0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,
	0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,
	0x0c,0x0c,0x13,0x13,0x13,0x0c,0x0c,0x0c,0x0c,0x13,0x13,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0e,0x0d,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0f,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
	0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,
};
